/*************************************************************************
 *
 * HTCd - Copyright (C) 1998-2006 Henrik Rydberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <http/http.h>

httpResp::httpResp()
{
  server="HTCd/1.2 MT (Unix)";
  version="HTTP/1.1";
  status=HTTP_200;
  mime_version.clear();
  //accept="*/*";
  //this is non-standard in response headers - now...
}

int httpResp::Special(sref line)
{
  version=First(line);
  status=Rest(line);
  return 1;
}

int httpResp::Handle(const sref& name,
                     const sref& rawval,
                     const sref& val,
                     const rrpile& opts)
{
  if(HTTP::Handle(name,rawval,val,opts)) return 1;
  else if(eqn(name,age_tag)) age=val;
  else if(eqn(name,etag_tag)) etag=val;
  else if(eqn(name,location_tag)) location=val;
  else if(eqn(name,proxy_authent_tag)) proxy_authent=val;
  else if(eqn(name,retry_tag)) retry=val;
  else if(eqn(name,server_tag)) server=val;
  else if(eqn(name,vary_tag)) vary=val;
  else if(eqn(name,authent_tag)) authent=val;
  else if(eqn(name,set_cookie_tag)) {
    for(rrpile::const_iterator p=opts.begin();p!=opts.end();p++)
      set_cookies.push_back(cookie_t(p->first,p->second));
  }
  else return 0;
  return 1;
}

void httpResp::putHead(mstream& out) const
{
  out<<version<<" "<<status<<CRLF;
  if(putval(out,age_tag,age)) putend(out);
  if(putval(out,etag_tag,etag)) putend(out);
  if(putval(out,location_tag,location)) putend(out);
  if(putval(out,proxy_authent_tag,proxy_authent)) putend(out);
  if(putval(out,retry_tag,retry)) putend(out);
  if(putval(out,server_tag,server)) putend(out);
  if(putval(out,vary_tag,vary)) putend(out);
  if(putval(out,authent_tag,authent)) putend(out);
  HTTP::putHead(out);
  for(msvec<cookie_t>::const_iterator p=set_cookies.begin();p!=set_cookies.end();p++) {
    out<<set_cookie_tag<<": ";
    putopt1(out,p->name,p->val);
    if(p->options.nempty()) out<<"; "<<p->options;
    putend(out);
  }
}
