/*************************************************************************
 *
 * HTCd - Copyright (C) 1998-2006 Henrik Rydberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef LOCKH
#define LOCKH

#include <mt/mstring.h>

// OFFERS TWO LOCKING MECHANISMS

void LockFile(const mstring& path,int prot=0660) throw(merror_t);
void UnlockFile(const mstring& path);

// These functions create an exclusive lock on path by
// creating a file path.lock.
// Works well and is the same mechnism used by
// mailing systems.

////////////////////////////////////////////

struct proclock_t {
  pid_t pid;
  mstring path;
  proclock_t(const mstring& p,int prot=0660);
  ~proclock_t();
};

#define PROCLOCK(p) proclock_t proclock(p)

// The proclock class creates a file named path.lock
// and stores the curent process id in it.
// once the destructor is called in the same
// process, the file is removed.

////////////////////////////////////////////

struct nodelock_t {
  pid_t pid;
  mstring path;
  nodelock_t(const mstring& p,int prot=0660);
  ~nodelock_t();
};

#define NODELOCK(p) nodelock_t nodelock(p)

// The nodelock class creates a file named path.lock
// and stores the curent network node name in it.
// once the destructor is called in the same
// process, the file is removed.

////////////////////////////////////////////

void ProcTerminate(const mstring& path,int stime=2);

#endif

