/*************************************************************************
 *
 * HTCd - Copyright (C) 1998-2006 Henrik Rydberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef UNIHEAPH
#define UNIHEAPH

#include <mt/mlock.h>
#include <malloc.h>
#include <string.h>

// NOTE - must be in extern/static scope, or zeroed before use
// by a call to init. constructors will fail badly.
// A minimum of sizeof(T)==4 is required.
// Furthermore, the upper limit of heap space is 1Gb.

template <int N> class uniheap {
public:
  static const int maxchunks=24;

  void init() {
    head=0;
    chunk=0;
    nchunks=0;
    memset(&mutex,0,sizeof(mutex_t));
  }

  char* push() { MLOCK(mutex); char* p=news(); head=*(char**)p; return p; }
  void pop(void* p) { MLOCK(mutex); *((char**)p)=head; head=(char*)p; }  
  int chunksize() const { return chunk; }
protected:
  char* news() {
    if(!head) {
      if(!chunk) chunk=N<<4;
      if(nchunks>=maxchunks) throw((int)0);
      head=chunks[nchunks++]=(char*)malloc(chunk);
      if(!head) throw((int)0);
      for(int i=0;i<chunk-N;i+=N) *((char**)(head+i))=head+i+N;
      *((char**)(head+chunk-N))=0;
      chunk<<=1;
    }
    return head;
  }
private:
  char *head,*chunks[maxchunks];
  int chunk,nchunks;
  mutex_t mutex;
};

#endif
