/*************************************************************************
 *
 * HTCd - Copyright (C) 1998-2006 Henrik Rydberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <sh/htc.h>
#include <hdb/buffer.h>
#include <mt/lock.h>

const mstring RCPATH=htchome()+sref("/htc/htc.rc");
const mstring left="{",right="}";
static fp_stream mout(stdout),merr(stderr);

static int unbalanced(const sref& in)
{
  static const charset tag("{[(");
  if(in.empty()) return 1;
  sref s=in.right_first_of(tag);
  while(s.nempty()) {
    int p=-1;
    switch(*s) {
    case '{': p=s.skip(cset_braces); break;
    case '[': p=s.skip(cset_indices); break;
    case '(': p=s.skip(cset_para); break;
    }
    if(p<0) return 1;
    s=s.adv(p).right_first_of(tag);
  }
  return 0;
}

main(int argc,char* argv[]) try
{
  NODELOCK(USERMAP);
  HTC htc(ms_empty);
  htc_user user(htc.toks);
  htc_req req(htc,user.toks); req.uri=Cwd();
  Env env(htc.ops,req,XSYSTEM);
  if(argc>1) {
    for(int i=2;i<argc;i++) env.xpush(left+sref(argv[i])+right);
    mstring buffer; LoadBuffer(buffer,argv[1]);
    if(*buffer=='#') buffer.erase(0,buffer.find('\n')+1);
    env.eval(mout,buffer,1);
  }
  else {
    mstring buffer,line;
    if(buffer.load(RCPATH.c_str())>=0) {
      Env e(env,user.toks,RCPATH);
      mstream m;
      e.parse(m,buffer);
    }
    while(!feof(stdin)) {
      merr<<"htc>";
      buffer.clear();
      while(!feof(stdin)&&unbalanced(buffer)) {
        getline(line,stdin);
        buffer.append(line);
        buffer+='\n';
      }
      try {
        env.parse(mout,buffer);
        mout<<"\n";
      }
      catch(const merror_t& e) {
        merr<<e.desc<<"\n";
      }
    }
  }
  return 0;
}
catch(const merror_t& e) {
  merr<<"ERROR: ["<<e.desc<<"]\n";
  return -1;
}
catch(...) {
  merr<<"ERROR: [FATAL]\n";
  return -1;
}
